# Copyright (c) 2021 SUSE LLC
#
# This file is part of azure_img_utils. azure_img_utils provides an
# api and command line utilities for handling images in the Azure Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import jmespath
import json
import re
import requests
import time

from datetime import date, datetime

from azure_img_utils.exceptions import AzureCloudPartnerException


def go_live_with_cloud_partner_offer(
    access_token: str,
    offer_id: str,
    publisher_id: str
) -> str:
    """
    Go live with cloud partner offer and return the operation location.
    """
    endpoint = get_cloud_partner_endpoint(
        offer_id,
        publisher_id,
        go_live=True
    )
    headers = get_cloud_partner_api_headers(
        access_token,
        content_type='application/json'
    )

    response = process_request(
        endpoint,
        headers,
        method='post',
        json_response=False
    )

    return response.headers['Location']


def get_cloud_partner_endpoint(
    offer_id: str,
    publisher_id: str,
    api_version: str = '2017-10-31',
    publish: bool = False,
    go_live: bool = False,
    status: bool = False
) -> str:
    """
    Return the endpoint URL to cloud partner API for offer and publisher.
    """
    endpoint = 'https://cloudpartner.azure.com/api/' \
               'publishers/{publisher_id}/' \
               'offers/{offer_id}' \
               '{method}' \
               '?api-version={api_version}'

    if publish:
        method = '/publish'
    elif go_live:
        method = '/golive'
    elif status:
        method = '/status'
    else:
        method = ''

    endpoint = endpoint.format(
        offer_id=offer_id,
        publisher_id=publisher_id,
        method=method,
        api_version=api_version
    )

    return endpoint


def get_cloud_partner_api_headers(
    access_token: str,
    content_type: str = None,
    if_match: str = None,
    content_length: str = None
) -> dict:
    """
    Return dictionary of request headers for cloud partner API.
    """
    headers = {
        'Accept': 'application/json',
        'Authorization': 'Bearer ' + access_token
    }

    if content_type:
        headers['Content-Type'] = content_type

    if if_match:
        headers['If-Match'] = if_match

    if content_length is not None:
        headers['Content-Length'] = content_length

    return headers


def get_cloud_partner_operation(access_token: str, operation: str) -> dict:
    """
    Get the status of the provided API operation.
    """
    endpoint = 'https://cloudpartner.azure.com{operation}'.format(
        operation=operation
    )

    headers = get_cloud_partner_api_headers(access_token)
    response = process_request(
        endpoint,
        headers
    )

    return response


def put_cloud_partner_offer_doc(
    access_token: str,
    doc: dict,
    offer_id: str,
    publisher_id: str
):
    """
    Put an updated cloud partner offer doc to the API.
    """
    endpoint = get_cloud_partner_endpoint(
        offer_id,
        publisher_id
    )
    headers = get_cloud_partner_api_headers(
        access_token,
        content_type='application/json',
        if_match='*'
    )

    response = process_request(
        endpoint,
        headers,
        data=doc,
        method='put'
    )

    return response


def publish_cloud_partner_offer(
    access_token: str,
    offer_id: str,
    publisher_id: str,
    notification_emails: str
):
    """
    Publish the cloud partner offer and return the operation location.
    """
    endpoint = get_cloud_partner_endpoint(
        offer_id,
        publisher_id,
        publish=True
    )
    headers = get_cloud_partner_api_headers(
        access_token,
        content_type='application/json'
    )

    response = process_request(
        endpoint,
        headers,
        data={'metadata': {'notification-emails': notification_emails}},
        method='post',
        json_response=False
    )

    return response.headers['Location']


def process_request(
    endpoint: str,
    headers: dict,
    data: dict = None,
    method: str = 'get',
    json_response: bool = True,
    retries: int = 5
):
    """
    Build and run API request.

    If the response code is not successful raise an exception for status.

    Return the response or json content.
    """
    kwargs = {
        'headers': headers
    }

    if data:
        kwargs['data'] = json.dumps(data)

    sleep = 1
    while True:
        try:
            response = getattr(requests, method)(
                endpoint,
                **kwargs
            )
        except requests.exceptions.ConnectionError:
            if retries <= 0:
                raise
            else:
                retries -= 1
                sleep = sleep * 2
                continue

        if response.status_code in (200, 202):
            break
        elif retries <= 0:
            response.raise_for_status()
        else:
            retries -= 1
            sleep = sleep * 2

        time.sleep(sleep)

    if json_response:
        return response.json()
    else:
        return response


def request_cloud_partner_offer_doc(
    access_token: str,
    offer_id: str,
    publisher_id: str
) -> dict:
    """
    Request a Cloud Partner Offer doc for the provided publisher and offer.
    """
    endpoint = get_cloud_partner_endpoint(
        offer_id,
        publisher_id
    )
    headers = get_cloud_partner_api_headers(access_token)

    response = process_request(
        endpoint,
        headers,
        method='get'
    )

    return response


def get_cloud_partner_offer_status(
    access_token: str,
    offer_id: str,
    publisher_id: str
) -> dict:
    """
    Returns the status of a Cloud Partner Offer based on id and publisher.

    If status is not found "unkown" is returned. If offer is publishing
    and publisher-signoff step is waitingForPublisherReview then this
    is the offer status. The offer is waiting for publisher to trigger
    go-live.
    """
    endpoint = get_cloud_partner_endpoint(
        offer_id,
        publisher_id,
        status=True
    )
    headers = get_cloud_partner_api_headers(access_token)

    response = process_request(
        endpoint,
        headers,
        method='get'
    )

    status = response.get('status', 'unkown')

    if status == 'running':
        signoff_status = jmespath.search(
            "steps[?stepName=='publisher-signoff'].status | [0]",
            response
        )
        if signoff_status == 'waitingForPublisherReview':
            status = 'waitingForPublisherReview'

    return status


def add_image_version_to_offer(
    doc: dict,
    blob_url: str,
    description: str,
    image_name: str,
    label: str,
    sku: str,
    vm_images_key: str = 'microsoft-azure-corevm.vmImagesPublicAzure',
    generation_id: str = None,
    cloud_image_name_generation_suffix: str = None
) -> dict:
    """
    Update the cloud partner offer doc with a new version of the given sku.
    """
    matches = re.findall(r'\d{8}', image_name)

    # If image name already has a date use it as release date.
    if matches:
        release_date = datetime.strptime(matches[0], '%Y%m%d').date()
    else:
        release_date = date.today()

    version = {
        'osVhdUrl': blob_url,
        'label': label,
        'mediaName': image_name,
        'publishedDate': release_date.strftime("%m/%d/%Y"),
        'description': description,
        'showInGui': True,
        'lunVhdDetails': []
    }

    for doc_sku in doc['definition']['plans']:
        if doc_sku['planId'] == sku:
            release = release_date.strftime("%Y.%m.%d")

            if vm_images_key not in doc_sku:
                doc_sku[vm_images_key] = {}

            doc_sku[vm_images_key][release] = version

            if generation_id:
                for plan in doc_sku['diskGenerations']:
                    if plan['planId'] == generation_id:
                        generation_version = copy.deepcopy(version)
                        generation_version['mediaName'] = '-'.join([
                            image_name,
                            cloud_image_name_generation_suffix or generation_id
                        ])

                        if vm_images_key not in plan:
                            plan[vm_images_key] = {}

                        plan[vm_images_key][release] = generation_version
                        break
                else:
                    raise AzureCloudPartnerException(
                        'No Match found for Generation ID: {gen}. '
                        'Offer doc not updated properly.'.format(
                            gen=generation_id
                        )
                    )

            break
    else:
        raise AzureCloudPartnerException(
            'No Match found for SKU: {sku}. '
            'Offer doc not updated properly.'.format(
                sku=sku
            )
        )

    return doc


def remove_image_version_from_offer(
    doc: dict,
    image_version: str,
    plan_id: str,
    vm_images_key: str = 'microsoft-azure-corevm.vmImagesPublicAzure'
) -> dict:
    """
    Remove the given image version from the cloud partner offer doc.
    """
    removed = False
    for doc_sku in doc['definition']['plans']:
        if doc_sku['planId'] == plan_id:
            if image_version in doc_sku[vm_images_key]:
                del doc_sku[vm_images_key][image_version]
                removed = True

        for plan in doc_sku['diskGenerations']:
            if plan['planId'] == plan_id:
                if image_version in plan[vm_images_key]:
                    del plan[vm_images_key][image_version]
                    removed = True

    if not removed:
        raise AzureCloudPartnerException(
            'No match found for version: {version} and Plan ID: {plan}.'
            ' Offer doc not updated properly.'.format(
                version=image_version,
                plan=plan_id
            )
        )

    return doc


def deprecate_image_in_offer_doc(
    doc: dict,
    image_name: str,
    sku: str,
    log_callback,
    vm_images_key: str = 'microsoft-azure-corevm.vmImagesPublicAzure'
) -> dict:
    """
    Deprecate the image in the cloud partner offer doc.

    The image is set to not show in gui.
    """
    matches = re.findall(r'\d{8}', image_name)

    if matches:
        release_date = datetime.strptime(matches[0], '%Y%m%d').date()
        release = release_date.strftime("%Y.%m.%d")
    else:
        # image name must have a date to generate release key
        return doc

    for doc_sku in doc['definition']['plans']:
        if doc_sku['planId'] == sku \
                and doc_sku.get(vm_images_key) \
                and doc_sku[vm_images_key].get(release):

            image = doc_sku[vm_images_key][release]

            if image['mediaName'] == image_name:
                image['showInGui'] = False
            else:
                log_callback(
                    'Deprecation image name, {0} does match the mediaName '
                    'attribute, {1}.'.format(
                        image_name,
                        image['mediaName']
                    )
                )

            break
    else:
        raise AzureCloudPartnerException(
            f'No Match found for image in the SKU: {sku}. '
            'Offer doc not updated properly.'
        )

    return doc



from unittest.mock import MagicMock, patch

from azure.storage.blob import BlobServiceClient
from azure.mgmt.storage import StorageManagementClient

from azure_img_utils.storage import (
    get_blob_service,
    get_blob_url,
    get_storage_account_key
)


class Key(object):
    def __init__(self, value) -> None:
        self.value = value


class AccountKeys(object):
    def __init__(self, key1, key2) -> None:
        self.keys = [Key(key1), Key(key2)]


creds = {'super': 'secret'}
sas_token = (
    'sp=rl&st=2021-08-27T20:23:21Z&se=2021-08-28T04:23:21Z'
    '&spr=https&sv=2020-08-04&sr=c&sig=supersecretstuffhere'
)


@patch('azure_img_utils.auth.generate_container_sas')
def test_get_blob_url(mock_generate_sas):
    bsc = MagicMock(spec=BlobServiceClient)
    bsc.credential = MagicMock()
    bsc.credential.account_key = 'supersecretstuffhere'

    mock_generate_sas.return_value = sas_token

    url = get_blob_url(
        bsc,
        'image_123.raw',
        'account',
        'images'
    )

    expected = (
        'https://account.blob.core.windows.net/images/image_123.raw?'
        'sp=rl&st=2021-08-27T20:23:21Z&se=2021-08-28T04:23:21Z&spr=https'
        '&sv=2020-08-04&sr=c&sig=supersecretstuffhere'
    )
    assert url == expected


@patch('azure_img_utils.storage.get_client_from_json')
def test_get_storage_account_key(mock_get_client):
    smc = MagicMock(spec=StorageManagementClient)
    smc.storage_accounts = MagicMock()
    smc.storage_accounts.list_keys.return_value = AccountKeys('123', '321')
    mock_get_client.return_value = smc

    result = get_storage_account_key(creds, 'group', 'account')
    assert result == '123'


@patch('azure_img_utils.storage.BlobServiceClient')
@patch('azure_img_utils.storage.get_storage_account_key')
def test_get_blob_service(mock_get_key, mock_blob_service):
    bsc = MagicMock(spec=BlobServiceClient)

    mock_get_key.return_value = '123'
    mock_blob_service.return_value = bsc

    # Get service from creds
    result = get_blob_service(creds, 'group', 'account')
    assert result == bsc

    # Get service from sas token
    result = get_blob_service(sas_token, 'account')
    assert result == bsc



from unittest.mock import MagicMock, patch
from requests import Response

from azure_img_utils.azure_image import AzureImage


class TestAzureCloudPartner(object):
    def setup_class(self):
        self.image = AzureImage(
            container='images',
            storage_account='account',
            credentials_file='tests/creds.json',
            resource_group='group'
        )

        # Mock access token
        self.image._access_token = 'supersecret'

    @patch('azure_img_utils.cloud_partner.process_request')
    def test_get_offer_doc(self, mock_process_request):
        mock_process_request.return_value = {'offer': 'doc'}
        doc = self.image.get_offer_doc('sles', 'suse')
        assert doc['offer'] == 'doc'

    @patch('azure_img_utils.cloud_partner.process_request')
    def test_upload_offer_doc(self, mock_process_request):
        doc = {'offer': 'doc'}
        self.image.upload_offer_doc('sles', 'suse', doc)

    @patch('azure_img_utils.cloud_partner.process_request')
    def test_add_image_to_offer(self, mock_process_request):
        doc = {
            'definition': {
                'plans': [
                    {
                        'planId': 'gen1',
                        'diskGenerations': [{'planId': 'image-gen2'}]
                    }
                ]
            }
        }

        mock_process_request.return_value = doc
        self.image.add_image_to_offer(
            'image.raw',
            'image123-v20111111',
            'This is a great image!',
            'sles',
            'suse',
            'suse-sles',
            'gen1',
            'bloburl',
            'image-gen2',
            'gen2'
        )

        vm_key = 'microsoft-azure-corevm.vmImagesPublicAzure'
        plan = doc['definition']['plans'][0]
        generation = plan['diskGenerations'][0][vm_key]['2011.11.11']

        assert plan['planId'] == 'gen1'
        assert generation['mediaName'] == 'image123-v20111111-gen2'
        assert generation['showInGui']

    @patch('azure_img_utils.cloud_partner.process_request')
    def test_remove_image_from_offer(self, mock_process_request):
        vm_key = 'microsoft-azure-corevm.vmImagesPublicAzure'
        doc = {
            'definition': {
                'plans': [
                    {
                        'planId': 'gen1',
                        vm_key: {'2011.11.11': {
                            'mediaName': 'image123-v20111111-gen2',
                            'showInGui': True
                        }},
                        'diskGenerations': [{
                            'planId': 'gen2',
                            vm_key: {'2011.11.11': {
                                'mediaName': 'image123-v20111111-gen2',
                                'showInGui': True
                            }}
                        }]
                    }
                ]
            }
        }

        mock_process_request.return_value = doc
        self.image.remove_image_from_offer('suse:sles:gen1:2011.11.11')
        self.image.remove_image_from_offer('suse:sles:gen2:2011.11.11')

        plan = doc['definition']['plans'][0]
        generations = plan['diskGenerations'][0][vm_key]

        assert '2011.11.11' not in plan[vm_key]
        assert '2011.11.11' not in generations

    @patch('azure_img_utils.cloud_partner.process_request')
    def test_publish_offer(self, mock_process_request):
        response = MagicMock(spec=Response)
        response.headers = {'Location': '/uri/to/operation/id'}
        mock_process_request.return_value = response

        operation = self.image.publish_offer('sles', 'suse', 'test@email.com')
        assert operation == '/uri/to/operation/id'

    @patch('azure_img_utils.cloud_partner.process_request')
    def test_go_live_with_offer(self, mock_process_request):
        response = MagicMock(spec=Response)
        response.headers = {'Location': '/uri/to/operation/id'}
        mock_process_request.return_value = response

        operation = self.image.go_live_with_offer('sles', 'suse')
        assert operation == '/uri/to/operation/id'

    @patch('azure_img_utils.cloud_partner.process_request')
    def test_get_offer_status(self, mock_process_request):
        mock_process_request.return_value = {
            'status': 'running',
            'steps': [{
                'stepName': 'publisher-signoff',
                'status': 'waitingForPublisherReview'
            }]
        }

        status = self.image.get_offer_status('sles', 'suse')
        assert status == 'waitingForPublisherReview'

    @patch('azure_img_utils.cloud_partner.process_request')
    def test_get_operation(self, mock_process_request):
        mock_process_request.return_value = {'operation': 'info'}
        operation = self.image.get_operation('/uri/to/operation/id')
        assert operation['operation'] == 'info'



# -*- coding: utf-8 -*-

"""Azure image utils exceptions module."""

# Copyright (c) 2021 SUSE LLC
#
# This file is part of azure_img_utils. azure_img_utils provides an
# api and command line utilities for handling images in the Azure Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class AzureImgUtilsException(Exception):
    """Generic exception for the azure_img_utils package."""


class AzureCloudPartnerException(AzureImgUtilsException):
    """Exception for Azure Cloud Partner processes."""


class AzureImgUtilsStorageException(AzureImgUtilsException):
    """Exception for Azure Storage processes."""



import pytest

from unittest.mock import MagicMock

from azure.mgmt.compute import ComputeManagementClient

from azure_img_utils.azure_image import AzureImage
from azure_img_utils.exceptions import AzureImgUtilsException


class Image(object):
    def __init__(self, name) -> None:
        self.name = name


class AsyncOperation(object):
    def result(self):
        pass


class TestAzureImageCompute(object):
    def setup_class(self):
        self.image = AzureImage(
            container='images',
            storage_account='account',
            credentials_file='tests/creds.json',
            resource_group='group'
        )

        # Mock compute client
        self.cc = MagicMock(spec=ComputeManagementClient)
        self.cc.images.list.return_value = [Image('test-image-123')]
        self.image._compute_client = self.cc

    def test_image_exists(self):
        assert self.image.image_exists('test-image-123')

    def test_get_compute_image(self):
        image = self.image.get_compute_image('test-image-123')
        assert image.name == 'test-image-123'

    def test_delete_compute_image(self):
        self.cc.images.begin_delete.return_value = AsyncOperation()
        self.image.delete_compute_image('test-image-123')
        assert self.cc.images.begin_delete.call_count == 1

    def test_create_compute_image(self):
        msg = 'Image already exists. To force deletion and re-create ' \
              'the image use "force_replace_image=True".'

        with pytest.raises(AzureImgUtilsException, match=msg):
            self.image.create_compute_image(
                'image_123.raw',
                'test-image-123',
                'southcentralus'
            )

        image_name = self.image.create_compute_image(
            'image_123.raw',
            'test-image-123',
            'southcentralus',
            force_replace_image=True
        )

        assert image_name == 'test-image-123'



# Copyright (c) 2021 SUSE LLC
#
# This file is part of azure_img_utils. azure_img_utils provides an
# api and command line utilities for handling images in the Azure Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import lzma
import subprocess


class FileType(object):
    """
    map file magic information to type methods
    """
    def __init__(self, file_name: str):
        self.file_name = file_name
        file_info = subprocess.Popen(
            ['file', file_name],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        self.filetype = file_info.communicate()[0].decode()

    def is_xz(self):
        if re.match('.*: XZ compressed', self.filetype):
            return True
        return False

    def get_size(self):
        if self.is_xz():
            with lzma.open(self.file_name) as lzma_stream:
                lzma_stream.seek(0, os.SEEK_END)
                return lzma_stream.tell()
        else:
            return os.path.getsize(self.file_name)



# Copyright (c) 2021 SUSE LLC
#
# This file is part of azure_img_utils. azure_img_utils provides an
# api and command line utilities for handling images in the Azure Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import msal

from datetime import datetime, timedelta

from azure.identity import ClientSecretCredential
from azure.storage.blob import (
    generate_container_sas,
    ContainerSasPermissions
)
from azure_img_utils.exceptions import AzureImgUtilsException


def acquire_access_token(credentials: dict, cloud_partner: bool = False):
    """
    Get an access token from msal library.

    credentials:
      A service account json dictionary.
    """
    client = msal.ConfidentialClientApplication(
        client_id=credentials.get('clientId'),
        client_credential=credentials.get('clientSecret'),
        authority='/'.join([
            credentials.get('activeDirectoryEndpointUrl'),
            credentials.get('tenantId')
        ])
    )

    if cloud_partner:
        resource = 'https://cloudpartner.azure.com/.default'
    else:
        resource = credentials['managementEndpointUrl'] + '.default'

    response = client.acquire_token_for_client(
        resource
    )

    if 'error' in response:
        raise AzureImgUtilsException(
            f'Unable to authenticate against {resource}: '
            f'{response.get("error")}'
        )

    return response.get('access_token')


def create_sas_token(
    blob_service,
    storage_account: str,
    container: str,
    permissions=ContainerSasPermissions(read=True, list=True),
    expire_hours: int = 1,
    start_hours: int = 1
):
    """
    Create a sas token for the given container.

    By default make the token valid 1 hour in the past and
    1 hour in the future with read and list permissions.
    """
    expiry_time = datetime.utcnow() + timedelta(hours=expire_hours)
    start_time = datetime.utcnow() - timedelta(hours=start_hours)

    return generate_container_sas(
        storage_account,
        container,
        permission=permissions,
        expiry=expiry_time,
        start=start_time,
        account_key=blob_service.credential.account_key
    )


def get_client_from_json(client, credentials: dict):
    """Return an authenticated client with the provided credentials."""
    credential = get_secret_credential(credentials)
    return client(
        credential,
        credentials['subscriptionId']
    )


def get_secret_credential(credentials: dict):
    """Create credentials object from credentials dictionary."""
    return ClientSecretCredential(
        tenant_id=credentials['tenantId'],
        client_id=credentials['clientId'],
        client_secret=credentials['clientSecret']
    )



import pytest

from unittest.mock import MagicMock

from azure.core.exceptions import ResourceNotFoundError
from azure.storage.blob import BlobServiceClient
from azure.storage.blob._container_client import ContainerClient
from azure.storage.blob._blob_client import BlobClient

from azure_img_utils.azure_image import AzureImage
from azure_img_utils.exceptions import (
    AzureImgUtilsException,
    AzureImgUtilsStorageException
)


class TestAzureImageStorage(object):
    def setup_class(self):
        self.image = AzureImage(
            container='images',
            storage_account='account',
            credentials_file='tests/creds.json',
            resource_group='group'
        )

        # Mock blob service client
        self.bsc = MagicMock(spec=BlobServiceClient)
        cc = MagicMock(spec=ContainerClient)
        self.bc = MagicMock(spec=BlobClient)

        cc.get_blob_client.return_value = self.bc
        self.bsc.get_container_client.return_value = cc
        self.image._blob_service_client = self.bsc

    def test_blob_exists(self):
        self.bc.exists.return_value = True
        assert self.image.image_blob_exists('blob123')

    def test_delete_blob_exception(self):
        self.bc.delete_blob.side_effect = ResourceNotFoundError('Not found!')
        assert self.image.delete_storage_blob('not_a_blob.txt') is False

        self.bc.delete_blob.side_effect = None
        assert self.image.delete_storage_blob('blob.txt')

    def test_upload_blob(self):
        self.bc.exists.return_value = True

        # Blob exists and no force replace
        with pytest.raises(Exception):
            self.image.upload_image_blob('tests/image.raw')

        blob = self.image.upload_image_blob(
            'tests/image.raw',
            force_replace_image=True,
            max_retry_attempts=5,
            max_workers=10
        )

        assert blob == 'image.raw'

    def test_upload_blob_exception(self):
        self.bc.exists.return_value = False

        # Blob upload file not found

        msg = 'Image file tests/not_a_image.raw not found. ' \
              'Ensure the path to the file is correct.'

        with pytest.raises(AzureImgUtilsException, match=msg):
            self.image.upload_image_blob('tests/not_a_image.raw')

        self.bc.upload_blob.side_effect = Exception('Permission denied')

        # Blob upload fails

        msg = 'Unable to upload tests/image.raw: Permission denied'

        with pytest.raises(AzureImgUtilsStorageException, match=msg):
            self.image.upload_image_blob('tests/image.raw')



# -*- coding: utf-8 -*-

"""Azure image class module."""

# Copyright (c) 2021 SUSE LLC
#
# This file is part of azure_img_utils. azure_img_utils provides an
# api and command line utilities for handling images in the Azure Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os

from azure.core.exceptions import ResourceNotFoundError
from azure.mgmt.compute import ComputeManagementClient

from azure_img_utils.auth import get_client_from_json, acquire_access_token
from azure_img_utils.exceptions import AzureImgUtilsException
from azure_img_utils.storage import (
    get_blob_service,
    get_blob_url,
    blob_exists,
    delete_blob,
    upload_azure_file
)
from azure_img_utils.compute import (
    create_image,
    create_gallery_image_definition_version,
    delete_image,
    get_image,
    image_exists,
    remove_gallery_image_version,
    retrieve_gallery_image_version
)
from azure_img_utils.cloud_partner import (
    get_cloud_partner_offer_status,
    get_cloud_partner_operation,
    request_cloud_partner_offer_doc,
    add_image_version_to_offer,
    put_cloud_partner_offer_doc,
    publish_cloud_partner_offer,
    go_live_with_cloud_partner_offer,
    remove_image_version_from_offer
)


class AzureImage(object):
    """
    Provides methods for handling compute images in Azure.
    """

    def __init__(
        self,
        container: str = None,
        storage_account: str = None,
        credentials: dict = None,
        credentials_file: str = None,
        resource_group: str = None,
        sas_token: str = None,
        log_level=logging.INFO,
        log_callback=None,
        timeout: int = 180
    ):
        """Initialize class and setup logging."""
        self.container = container
        self.timeout = timeout
        self._blob_service_client = None
        self._compute_client = None
        self._access_token = None
        self._credentials = credentials
        self._credentials_file = credentials_file
        self._resource_group = resource_group
        self._storage_account = storage_account
        self._sas_token = sas_token

        if log_callback:
            self.log = log_callback
        else:
            self.log = logging.getLogger('azure-img-utils')
            self.log.setLevel(log_level)

        try:
            self.log_level = self.log.level
        except AttributeError:
            self.log_level = self.log.logger.level  # LoggerAdapter

    def image_blob_exists(self, blob_name: str):
        """Return True if image blob exists in the configured container."""
        return blob_exists(
            self.blob_service_client,
            blob_name,
            self.container
        )

    def delete_storage_blob(self, blob_name: str):
        """Delete blob if it exists in the configured container."""
        try:
            delete_blob(
                self.blob_service_client,
                blob_name,
                self.container
            )
        except ResourceNotFoundError:
            self.log.debug(
                f'Blob {blob_name} not found. '
                f'Nothing has been deleted.'
            )
            return False

        return True

    def upload_image_blob(
        self,
        image_file: str,
        max_workers: int = None,
        max_retry_attempts: int = None,
        blob_name: str = None,
        force_replace_image: bool = False,
        is_page_blob: bool = True,
        expand_image: bool = True
    ):
        """
        Upload image tarball to the configured container.

        Generate blob name based on image file path if a
        name is not provided.
        """
        if not blob_name:
            blob_name = image_file.rsplit(os.sep, maxsplit=1)[-1]

        if self.image_blob_exists(blob_name) and not force_replace_image:
            raise Exception(
                f'Image {blob_name} already exists. To replace an existing '
                f'image use force_replace_image option.'
            )
        elif self.image_blob_exists(blob_name) and force_replace_image:
            self.delete_storage_blob(blob_name)

        kwargs = {
            'is_page_blob': is_page_blob,
            'expand_image': expand_image
        }

        if max_workers:
            kwargs['max_workers'] = max_workers

        if max_retry_attempts:
            kwargs['max_retry_attempts'] = max_retry_attempts

        try:
            upload_azure_file(
                blob_name,
                self.container,
                image_file,
                self.blob_service_client,
                **kwargs
            )
        except FileNotFoundError:
            raise AzureImgUtilsException(
                f'Image file {image_file} not found. Ensure the path to'
                f' the file is correct.'
            )

        return blob_name

    def image_exists(self, image_name: str) -> bool:
        """Return True if image exists, false otherwise."""
        return image_exists(self.compute_client, image_name)

    def gallery_image_version_exists(
        self,
        gallery_name: str,
        gallery_image_name: str,
        image_version: str
    ) -> bool:
        """
        Return True if gallery image version exists, false otherwise.
        """
        gallery_image_version = self.get_gallery_image_version(
            gallery_name,
            gallery_image_name,
            image_version
        )
        return gallery_image_version is not None

    def delete_compute_image(self, image_name: str):
        """
        Delete compute image.
        """
        if not self.resource_group:
            raise AzureImgUtilsException(
                'Resource group is required to delete a compute image'
            )

        delete_image(
            self.compute_client,
            self.resource_group,
            image_name
        )

    def delete_gallery_image_version(
        self,
        gallery_name: str,
        gallery_image_name: str,
        image_version: str
    ):
        """
        Delete gallery image version.
        """
        if not self.resource_group:
            raise AzureImgUtilsException(
                'Resource group is required to delete a gallery image'
            )

        remove_gallery_image_version(
            gallery_name,
            gallery_image_name,
            image_version,
            self.resource_group,
            self.compute_client
        )

    def get_compute_image(self, image_name: str) -> dict:
        """
        Return compute image by name.

        If image is not found None is returned.
        """
        return get_image(self.compute_client, image_name)

    def get_gallery_image_version(
        self,
        gallery_name: str,
        gallery_image_name: str,
        image_version: str
    ) -> dict:
        """
        Return gallery image by gallery_image_name.

        If image is not found None is returned.
        """
        if not self.resource_group:
            raise AzureImgUtilsException(
                'Resource group is required to retrieve a gallery image'
            )

        return retrieve_gallery_image_version(
            gallery_name,
            gallery_image_name,
            image_version,
            self.resource_group,
            self.compute_client
        )

    def create_compute_image(
        self,
        blob_name: str,
        image_name: str,
        region: str,
        force_replace_image: bool = False,
        hyper_v_generation: str = 'V1'
    ) -> str:
        """
        Create compute image from storage blob.

        If image exists and force replace is True delete
        the existing image before creation.
        """
        if not self.container:
            raise AzureImgUtilsException(
                'Container is required to create a compute image'
            )

        if not self.resource_group:
            raise AzureImgUtilsException(
                'Resource group is required to create a compute image'
            )

        if not self.storage_account:
            raise AzureImgUtilsException(
                'Storage account is required to create a compute image'
            )

        exists = image_exists(self.compute_client, image_name)

        if exists and force_replace_image:
            delete_image(
                self.compute_client,
                self.resource_group,
                image_name
            )
        elif exists and not force_replace_image:
            raise AzureImgUtilsException(
                'Image already exists. To force deletion and re-create '
                'the image use "force_replace_image=True".'
            )

        return create_image(
            blob_name,
            image_name,
            self.compute_client,
            self.container,
            self.resource_group,
            self.storage_account,
            region,
            hyper_v_generation
        )

    def create_gallery_image_version(
        self,
        blob_name: str,
        gallery_name: str,
        gallery_image_name: str,
        image_version: str,
        region: str,
        force_replace_image: bool = False
    ) -> str:
        """
        Create gallery image version from storage blob.

        If gallery image version exists and force replace is True
        then delete the existing image before creation. Otherwise
        an error is raised.
        """
        if not self.container:
            raise AzureImgUtilsException(
                'Container is required to create a gallery image'
            )

        if not self.resource_group:
            raise AzureImgUtilsException(
                'Resource group is required to create a gallery image'
            )

        if not self.storage_account:
            raise AzureImgUtilsException(
                'Storage account is required to create a gallery image'
            )

        exists = self.gallery_image_version_exists(
            gallery_name,
            gallery_image_name,
            image_version
        )

        if exists and force_replace_image:
            self.delete_gallery_image_version(
                gallery_name,
                gallery_image_name,
                image_version
            )
        elif exists and not force_replace_image:
            raise AzureImgUtilsException(
                'Gallery image version already exists. To force deletion and '
                're-create the image set "force_replace_image" to True.'
            )

        return create_gallery_image_definition_version(
            blob_name,
            gallery_name,
            gallery_image_name,
            image_version,
            self.resource_group,
            region,
            self.storage_account,
            self.container,
            self.compute_client
        )

    def get_offer_doc(
        self,
        offer_id: str,
        publisher_id: str
    ) -> dict:
        """
        Return the offer doc dictionary for the given offer.
        """
        return request_cloud_partner_offer_doc(
            self.access_token,
            offer_id,
            publisher_id
        )

    def upload_offer_doc(
        self,
        offer_id: str,
        publisher_id: str,
        offer_doc: dict
    ):
        """
        Upload the offer doc for the given offer.

        offer_doc is a dictionary defining the offer details.
        """
        put_cloud_partner_offer_doc(
            self.access_token,
            offer_doc,
            offer_id,
            publisher_id
        )

    def add_image_to_offer(
        self,
        blob_name: str,
        image_name: str,
        image_description: str,
        offer_id: str,
        publisher_id: str,
        label: str,
        sku: str,
        blob_url: str = None,
        generation_id: str = None,
        generation_suffix: str = None,
        vm_images_key: str = None
    ):
        """
        Add a new image version to the given offer.

        The offer is pulled from the partner center, updated with the
        new image version and re-uploaded. To make the new image available
        the offer must be published and set to go-live.

        A blob_url is generated for the container if one is not provided.
        """
        if not self.container:
            raise AzureImgUtilsException(
                'Container is required to add an image to an offer'
            )

        if not blob_url:
            blob_url = get_blob_url(
                self.blob_service_client,
                blob_name,
                self.storage_account,
                self.container,
                expire_hours=24 * 92,
                start_hours=24
            )

        offer_doc = self.get_offer_doc(offer_id, publisher_id)

        kwargs = {
            'generation_id': generation_id,
            'cloud_image_name_generation_suffix': generation_suffix
        }

        if vm_images_key:
            kwargs['vm_images_key'] = vm_images_key

        offer_doc = add_image_version_to_offer(
            offer_doc,
            blob_url,
            image_description,
            image_name,
            label,
            sku,
            **kwargs
        )
        self.upload_offer_doc(
            offer_id,
            publisher_id,
            offer_doc
        )

    def remove_image_from_offer(
        self,
        image_urn: str,
        vm_images_key: str = None
    ):
        """
        Delete the given image version from the offer.

        The offer is pulled from the partner center, the old image version
        is deleted and re-uploaded. To make the new image available
        the offer must be published and set to go-live.
        """
        publisher_id, offer_id, plan_id, image_version = image_urn.split(':')
        offer_doc = self.get_offer_doc(offer_id, publisher_id)

        kwargs = {}
        if vm_images_key:
            kwargs['vm_images_key'] = vm_images_key

        offer_doc = remove_image_version_from_offer(
            offer_doc,
            image_version,
            plan_id,
            **kwargs
        )
        self.upload_offer_doc(
            offer_id,
            publisher_id,
            offer_doc
        )

    def publish_offer(
        self,
        offer_id: str,
        publisher_id: str,
        notification_emails: str
    ) -> str:
        """
        Publish the given offer.

        notification_emails is required to be a comma separated list
        of emails. This argument is required. However, for migrated
        offers the emails are ignored. For migrated offers
        notifications will be sent to the email address set in the
        Seller contact info section of your Account settings in
        Partner Center.

        Returns the operation uri.
        """
        return publish_cloud_partner_offer(
            self.access_token,
            offer_id,
            publisher_id,
            notification_emails
        )

    def go_live_with_offer(
        self,
        offer_id: str,
        publisher_id: str
    ) -> str:
        """
        Set the offer as go-live.

        This makes all new changes to the offer publicly visible.

        Returns the operation uri.
        """
        return go_live_with_cloud_partner_offer(
            self.access_token,
            offer_id,
            publisher_id
        )

    def get_offer_status(self, offer_id, publisher_id) -> str:
        """
        Returns the status of the offer.
        """
        return get_cloud_partner_offer_status(
            self.access_token,
            offer_id,
            publisher_id
        )

    def get_operation(self, operation: str) -> dict:
        """
        Returns a dictionary status for the given operation.
        """
        return get_cloud_partner_operation(self.access_token, operation)

    @property
    def blob_service_client(self):
        """
        Blob service client property

        Lazy blob service client initialization.
        """
        if not self.storage_account:
            raise AzureImgUtilsException(
                'Storage account is required to authenticate storage '
                'blob operations.'
            )

        if not self._blob_service_client:
            if self.sas_token:
                args = (
                    self.sas_token,
                    self.storage_account
                )
            elif self.credentials and self.resource_group:
                args = (
                    self.credentials,
                    self.resource_group,
                    self.storage_account
                )
            else:
                raise Exception(
                    'Either an sas_token or credentials_file/credentials and '
                    'resource_group is required to authenticate any '
                    'operations.'
                )

            self._blob_service_client = get_blob_service(*args)

        return self._blob_service_client

    @property
    def compute_client(self):
        """
        Lazy compute client attribute

        If compute client is not set create a new client from credentials.
        """
        if not self._compute_client:
            self._compute_client = get_client_from_json(
                ComputeManagementClient,
                self.credentials
            )

        return self._compute_client

    @property
    def access_token(self):
        if not self._access_token:
            self._access_token = acquire_access_token(
                self.credentials,
                cloud_partner=True
            )

        return self._access_token

    @property
    def credentials(self):
        """
        Lazy credentials attribute

        If credentials not set and a file is available attempt
        to load credentials json as dictionary.
        """
        if not self._credentials and self._credentials_file:
            creds_file = os.path.expanduser(self._credentials_file)

            with open(creds_file, 'r') as json_file:
                self._credentials = json.load(json_file)

        return self._credentials

    @credentials.setter
    def credentials(self, creds):
        """
        Invalidates the blob service and compute clients.
        """
        self._credentials = creds
        self._blob_service_client = None
        self._compute_client = None

    @property
    def credentials_file(self):
        return self._credentials_file

    @credentials_file.setter
    def credentials_file(self, creds_file):
        """
        Invalidates the credentials.
        """
        self._credentials_file = creds_file
        self.credentials = None

    @property
    def sas_token(self):
        return self._sas_token

    @sas_token.setter
    def sas_token(self, token):
        """
        Invalidates the blob service client.
        """
        self._sas_token = token
        self._blob_service_client = None

    @property
    def resource_group(self):
        return self._resource_group

    @resource_group.setter
    def resource_group(self, group):
        """
        Invalidates the blob service client.
        """
        self._resource_group = group
        self._blob_service_client = None

    @property
    def storage_account(self):
        return self._storage_account

    @storage_account.setter
    def storage_account(self, account):
        """
        Invalidates the blob service client.
        """
        self._storage_account = account
        self._blob_service_client = None



# Copyright (c) 2021 SUSE LLC
#
# This file is part of azure_img_utils. azure_img_utils provides an
# api and command line utilities for handling images in the Azure Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


def create_image(
    blob_name: str,
    image_name: str,
    compute_client,
    container: str,
    resource_group: str,
    storage_account: str,
    region: str,
    hyper_v_generation: str = 'V1'
):
    """
    Create image in ARM from existing page blob.

    hyper v generation of V2 is uefi and V1 is legacy bios.
    """
    async_create_image = compute_client.images.begin_create_or_update(
        resource_group,
        image_name, {
            'location': region,
            'hyper_v_generation': hyper_v_generation,
            'storage_profile': {
                'os_disk': {
                    'os_type': 'Linux',
                    'os_state': 'Generalized',
                    'caching': 'ReadWrite',
                    'blob_uri': 'https://{0}.{1}/{2}/{3}'.format(
                        storage_account,
                        'blob.core.windows.net',
                        container,
                        blob_name
                    )
                }
            }
        }
    )
    async_create_image.result()

    return image_name


def create_gallery_image_definition_version(
    blob_name: str,
    gallery_name: str,
    gallery_image_name: str,
    image_version: str,
    gallery_resource_group: str,
    region: str,
    storage_account: str,
    container: str,
    compute_client
):
    """
    Create new gallery image definition version

    The image is replicated only in one source region.
    """
    subscription_id = compute_client._config.subscription_id
    image_profile = {
        'location': region,
        'publishing_profile': {
            'target_regions': [
                {
                    'name': region
                }
            ]
        },
        'storage_profile': {
            'os_disk_image': {
                'source': {
                    'id': f'/subscriptions/{subscription_id}/'
                          f'resourceGroups/{gallery_resource_group}/'
                          'providers/Microsoft.Storage/'
                          f'storageAccounts/{storage_account}',
                    'uri': f'https://{storage_account}.blob.core.windows.net/'
                           f'{container}/{blob_name}'
                },
                'host_caching': 'ReadWrite'
            }
        }
    }

    async_create = compute_client.gallery_image_versions.begin_create_or_update(  # noqa
        gallery_resource_group,
        gallery_name,
        gallery_image_name,
        image_version,
        image_profile
    )
    async_create.result()

    return gallery_image_name


def delete_image(compute_client, resource_group: str, image_name: str):
    """
    Delete the image from resource group.
    """
    async_delete_image = compute_client.images.begin_delete(
        resource_group,
        image_name
    )
    async_delete_image.result()


def remove_gallery_image_version(
    gallery_name: str,
    gallery_image_name: str,
    image_version: str,
    gallery_resource_group: str,
    compute_client
):
    """
    Delete the gallery image version from the gallery definition.
    """
    async_delete_image = compute_client.gallery_image_versions.begin_delete(
        gallery_resource_group,
        gallery_name,
        gallery_image_name,
        image_version
    )
    async_delete_image.result()


def list_images(compute_client):
    """
    Returns a list of image names.
    """
    images = compute_client.images.list()

    names = []
    for image in images:
        names.append(image.name)

    return names


def image_exists(compute_client, image_name: str):
    """
    Return True if an image with name image_name exists.
    """
    images = list_images(compute_client)
    return image_name in images


def get_image(compute_client, image_name: str):
    """
    Return image if it exists based on the image name.
    """
    images = compute_client.images.list()

    for image in images:
        if image_name == image.name:
            return image


def retrieve_gallery_image_version(
    gallery_name: str,
    gallery_image_name: str,
    image_version: str,
    gallery_resource_group: str,
    compute_client
):
    """
    Return gallery image if it exists based on the gallery_image_name.

    Return None if an exception is raised. A generic exception is raised
    if the resource is not found.
    """
    try:
        image = compute_client.gallery_image_versions.get(
            gallery_resource_group,
            gallery_name,
            gallery_image_name,
            image_version
        )
    except Exception:
        # A generic exception is always raised if resource not found
        # so there's no disadvantage to return None.
        return None

    return image.as_dict()



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""Setup script."""

# Copyright (c) 2021 SUSE LLC
#
# This file is part of azure_img_utils. azure_img_utils provides an
# api and command line utilities for handling images in the Azure Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import find_packages, setup

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('requirements.txt') as req_file:
    requirements = req_file.read().splitlines()

with open('requirements-test.txt') as req_file:
    test_requirements = req_file.read().splitlines()[2:]

with open('requirements-dev.txt') as req_file:
    dev_requirements = test_requirements + req_file.read().splitlines()[2:]


setup(
    name='azure-img-utils',
    version='0.4.0',
    description='Package that provides utilities for '
                'handling images in Azure Cloud.',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='SUSE',
    author_email='public-cloud-dev@susecloud.net',
    url='https://github.com/SUSE-Enceladus/azure-img-utils',
    packages=find_packages(),
    package_dir={
        'azure_img_utils': 'azure_img_utils'
    },
    entry_points={
        'console_scripts': [
            'azure-img-utils=azure_img_utils.azure_cli:main'
        ]
    },
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=requirements,
    extras_require={
        'dev': dev_requirements,
        'test': test_requirements
    },
    license='GPLv3+',
    zip_safe=False,
    keywords='azure-img-utils azure_img_utils',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: System :: Software Distribution',
        'License :: OSI Approved :: '
        'GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



# Copyright (c) 2021 SUSE LLC
#
# This file is part of azure_img_utils. azure_img_utils provides an
# api and command line utilities for handling images in the Azure Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import lzma

from functools import singledispatch

from azure.mgmt.storage import StorageManagementClient
from azure.storage.blob import (
    BlobServiceClient,
    ContainerSasPermissions
)

from azure_img_utils.auth import create_sas_token, get_client_from_json
from azure_img_utils.exceptions import AzureImgUtilsStorageException
from azure_img_utils.filetype import FileType


def delete_blob(
    blob_service_client,
    blob: str,
    container: str
):
    """
    Delete page blob in container.
    """
    blob_client = get_blob_client(blob_service_client, blob, container)
    blob_client.delete_blob()


def blob_exists(
    blob_service_client,
    blob: str,
    container: str
):
    """
    Return True if blob exists in container.
    """
    blob_client = get_blob_client(blob_service_client, blob, container)
    return blob_client.exists()


def get_blob_url(
    blob_service_client,
    blob_name: str,
    storage_account: str,
    container: str,
    permissions=ContainerSasPermissions(read=True, list=True),
    expire_hours: int = 1,
    start_hours: int = 1
):
    """
    Create a URL for the given blob with a shared access signature.

    The signature will expire based on expire_hours.
    """
    sas_token = create_sas_token(
        blob_service_client,
        storage_account,
        container,
        permissions=permissions,
        expire_hours=expire_hours,
        start_hours=start_hours
    )

    source_blob_url = (
        'https://{account}.blob.core.windows.net/'
        '{container}/{blob}?{token}'.format(
            account=storage_account,
            container=container,
            blob=blob_name,
            token=sas_token
        )
    )

    return source_blob_url


def get_storage_account_key(
    credentials: dict,
    resource_group: str,
    storage_account: str
):
    """Return the first storage account key for the provided account."""
    storage_client = get_client_from_json(
        StorageManagementClient,
        credentials
    )
    storage_key_list = storage_client.storage_accounts.list_keys(
        resource_group,
        storage_account
    )
    return storage_key_list.keys[0].value


def get_blob_client(blob_service_client, blob_name: str, container: str):
    """Return blob client based on container and blob name."""
    try:
        container_client = blob_service_client.get_container_client(container)
        blob_client = container_client.get_blob_client(blob_name)
    except ValueError as error:
        raise AzureImgUtilsStorageException(error) from error

    return blob_client


@singledispatch
def get_blob_service(
    credentials: dict,
    resource_group: str,
    storage_account: str
):
    """
    Return authenticated blob service instance for the storage account.

    Using storage account keys.
    """
    account_key = get_storage_account_key(
        credentials,
        resource_group,
        storage_account
    )

    return BlobServiceClient(
        account_url='https://{account_name}.blob.core.windows.net'.format(
            account_name=storage_account
        ),
        credential=account_key
    )


@get_blob_service.register(str)
def _(sas_token: str, storage_account: str):
    """
    Return authenticated page blob service instance for the storage account.

    Using an sas token.
    """
    return BlobServiceClient(
        account_url='https://{account_name}.blob.core.windows.net'.format(
            account_name=storage_account
        ),
        credential=sas_token
    )


def upload_azure_file(
    blob_name: str,
    container: str,
    file_name: str,
    blob_service_client,
    max_retry_attempts: int = 5,
    max_workers: int = 5,
    is_page_blob: bool = False,
    expand_image: bool = True
):
    """
    Upload file to Azure storage container.

    Authentication can be an sas token or with storage account
    keys.

    Blob can be block or page and if the blob is an image tarball
    it can be expanded during upload.
    """
    blob_client = get_blob_client(blob_service_client, blob_name, container)

    if is_page_blob:
        blob_type = 'PageBlob'
    else:
        blob_type = 'BlockBlob'

    system_image_file_type = FileType(file_name)
    if system_image_file_type.is_xz() and expand_image:
        open_image = lzma.LZMAFile
    else:
        open_image = open

    msg = ''
    while max_retry_attempts > 0:
        with open_image(file_name, 'rb') as image_stream:
            try:
                blob_client.upload_blob(
                    image_stream,
                    blob_type=blob_type,
                    length=system_image_file_type.get_size(),
                    max_concurrency=max_workers
                )
                return
            except Exception as error:
                msg = error
                max_retry_attempts -= 1

    raise AzureImgUtilsStorageException(
        'Unable to upload {0}: {1}'.format(
            file_name,
            msg
        )
    )



# -*- coding: utf-8 -*-

"""azure-img-utils Azure image utilities."""

# Copyright (c) 2021 SUSE LLC
#
# This file is part of azure_img_utils. azure_img_utils provides an
# api and command line utilities for handling images in the Azure Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__author__ = """SUSE"""
__email__ = 'public-cloud-dev@susecloud.net'
__version__ = '0.4.0'



import pytest

from unittest.mock import MagicMock

from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.compute._configuration import (
    ComputeManagementClientConfiguration
)

from azure_img_utils.azure_image import AzureImage
from azure_img_utils.exceptions import AzureImgUtilsException


class Image(object):
    def as_dict(self):
        return {
            'id': '123',
            'name': '2022.02.02',
            'type': 'Microsoft.Compute/galleries/images/versions',
            'location': 'westus2'
        }


class AsyncOperation(object):
    def result(self):
        pass


class TestAzureImageCompute(object):
    def setup_class(self):
        self.image = AzureImage(
            container='images',
            storage_account='account',
            credentials_file='tests/creds.json',
            resource_group='group'
        )

        # Mock compute client
        self.cc = MagicMock(spec=ComputeManagementClient)
        _config = MagicMock(spec=ComputeManagementClientConfiguration)
        _config.subscription_id = '123456789'
        self.cc._config = _config
        self.cc.gallery_image_versions.get.return_value = Image()
        self.image._compute_client = self.cc

    def test_gallery_image_version_exists(self):
        assert self.image.gallery_image_version_exists(
            'gallery1',
            'galleryimage1',
            '2022.02.02'
        )

    def test_get_gallery_image_version(self):
        image = self.image.get_gallery_image_version(
            'gallery1',
            'galleryimage1',
            '2022.02.02'
        )
        assert image['name'] == '2022.02.02'

    def test_delete_gallery_image_version(self):
        self.cc.gallery_image_versions.begin_delete.return_value = AsyncOperation()  # noqa
        self.image.delete_gallery_image_version(
            'gallery1',
            'galleryimage1',
            '2022.02.02'
        )
        assert self.cc.gallery_image_versions.begin_delete.call_count == 1

    def test_create_gallery_image_version(self):
        msg = 'Gallery image version already exists. To force deletion and ' \
              're-create the image set "force_replace_image" to True.'

        with pytest.raises(AzureImgUtilsException, match=msg):
            self.image.create_gallery_image_version(
                'image_123.vhd',
                'gallery1',
                'galleryimage1',
                '2022.02.02',
                'westus2',
            )

        image_name = self.image.create_gallery_image_version(
            'image_123.vhd',
            'gallery1',
            'galleryimage1',
            '2022.02.02',
            'westus2',
            force_replace_image=True
        )

        assert image_name == 'galleryimage1'
